unit grade;

interface

uses
	GeneralData, Windows, SysUtils, PsAPI;

var
	JobQueueBusy: Boolean;
    StopGrading: Boolean;
    GradeThreadExec: Boolean = false;
    indGrade: Integer;

procedure DoGrading(const Job: TJob);
procedure GradeThreadProc(lpData: Pointer);

implementation

uses Math;

{ FindTask()
  Function find's contestant program which will be graded }
function FindTask(Folder, Task: String): String;
const
	N = 5;
	Extension: array[1..N] of String = ('c', 'cpp', 'cc', 'pas', 'pp');
var
	i: Integer;
begin
	for i := 1 to N do
    begin
    	Result := Folder + Task + '.' + Extension[i];
    	if FileExists(Result) then
        	Exit;
    end;
    Result := '';
end;

{ Compile()
  Function compiles given task. Returns name of executable }
function Compile(const FileName, LogFile: String): String;
var
	f: TextFile;
    CompilerS, CmdLine, Executable: String;
    i: Integer;

    si: TStartupInfo;
	p: TProcessInformation;
begin
	Result := '';
    try
    	{ Read first line }
	    AssignFile(f, FileName);
	    Reset(f);
	    	ReadLn(f, CompilerS);
    	CloseFile(f);

        { Try to find fitting compiler }
        CmdLine := '';
        for i := 0 to Length( CompilerList )-1 do
            if pos(' ' + CompilerList[i].Sign + ' ', CompilerS) <> 0 then
            begin
            	CmdLine := CompilerList[i].CmdLine;
            	Break;
            end;

        if CmdLine = '' then
        begin
        	StoreToFile(LogFile, ' ( )  !');
            Abort;
        end;

        { Erase old executable }
        Executable := WorkDir + '\' + TEST_DIR + '\a.exe';
        DeleteFile(Executable);

        { Finish creating command line }
        ReplaceStg(CmdLine, '$file', GetName(FileName));
        ReplaceStg(CmdLine, '$exec', Executable);

		{ Launch compilation }
       	FillChar(si, sizeof(si), 0);
		with si do
        begin
			cb := sizeof(si);
			dwFlags := STARTF_USESHOWWINDOW;
			wShowWindow := SW_SHOWNOACTIVATE;
		end;

	    if not CreateProcess(NIL, PChar(CmdLine), NIL, NIL, FALSE,
                      CREATE_NO_WINDOW, NIL,
                      PChar(WorkDir + '\' + TEST_DIR), si, p) then
            begin
                StoreToFile(LogFile, '   !');
            	MessageBox(0, PChar('Error #'+ IntToStr(GetLastError)), NIL, 0);
                Abort;
            end;
		WaitForSingleObject(p.hProcess, INFINITE);

		if not FileExists(Executable) then
        	Abort else
            Result := Executable;
    except
    	Exit;
    end;
end;

function TestProcess(Hand: THandle; TimeLimit, MemoryLimit: DWORD): String;
var
    WaitCode, ExitCode: DWORD;
    CreationTime, ExitTime, KernelTime, UserTime: FILETIME;
    UserNano100: Int64 absolute UserTime;
    KernelNano100: Int64 absolute KernelTime;
    Time,Our_Time,Last_Time,Porog: Int64;
    MemoryData: PROCESS_MEMORY_COUNTERS;
    kill: Boolean;
begin
    MemoryData.cb := SizeOf(MemoryData);
    Result := '';

    kill := False;
    Our_Time := GetTickCount;
    Porog := 0;
    Last_Time := 0;
    repeat
        WaitCode := WaitForSingleObject(Hand, 0);
        if WaitCode = WAIT_FAILED then
        begin
            WriteLn('WaitForSingleObject() failed!');
            Result := 'WaitForSingleObject() failed!';
            Exit;
        end;

        GetProcessTimes(Hand, CreationTime, ExitTime, KernelTime, UserTime);
        Time := (UserNano100 div 10000) + (KernelNano100 div 10000);

        GetProcessMemoryInfo(Hand, @MemoryData, SizeOf(MemoryData));


        if MemoryData.PeakWorkingSetSize > MemoryLimit then
        begin
            Result := '   ''';
            kill := True;
            Break;
        end;
        if(Time = Last_Time) then
        begin
          Porog := Porog + (GetTickCount - Our_Time);
        end
        else Last_Time := Time;
        Our_Time := GetTickCount;

        if Porog > TimeLimit then
        begin
          Result := '    ( readln)';
          kill := True;
          break;
        end;
    until (Time > TimeLimit) or (WaitCode = WAIT_OBJECT_0);

    if (WaitCode <> WAIT_OBJECT_0) or (Time > TimeLimit) or kill then
        TerminateProcess(Hand, 0);

    if kill then
    begin
        sleep(500);
        Exit;
    end;

    if (Time > TimeLimit) then
    begin
        Result := '   ';
        Exit;
    end;

    if GetExitCodeProcess(Hand, ExitCode) then
	    if ExitCode <> 0 then
    	begin
        	Result := '   ';
	        Exit;
    	end;
end;

{ DoTest()
  This procedure does all grading work. Big thanks to D. Neiter for his
  help on this code ;). Thanks man! }
function DoTest(const Executable, FullLog, TaskName, TaskFolder: String;
			    iTest, maxScore, TimeLimit, MemoryLimit: Integer): Integer;
var
	ExecFolder, R, CallLine, TestNum: String;
    f, fRes: TextFile;
    StartInfo: TStartupInfo;
    ProcessInfo: TProcessInformation;
    i,ttt: Integer;
begin
	Result := 0;
	ExecFolder := Executable;
    while ExecFolder[ Length(ExecFolder) ] <> '\' do
    	Delete(ExecFolder, Length(ExecFolder), 1);

    { Open for logging }
    AssignFile(f, FullLog);
    Append(f);

    { Copy input file }
    TestNum := IntToStr(iTest);
    while Length( TestNum ) < 2 do
    	TestNum := '0' + TestNum;
    Sleep(100);
    if not CopyFile(PChar(TaskFolder + TestNum + '.in'),
                    PChar(ExecFolder + TaskName + '.in'),
                    False) then
    begin
    	WriteLn(f, TestNum, ')  0 -    ');
        CloseFile(f);
        Exit;
    end;

    { Remove output file }
    DeleteFile( PChar(ExecFolder + TaskName + '.out') );

    { Run contestant's program }
    FillChar(StartInfo, SizeOf(StartInfo), 0);
    StartInfo.cb := SizeOf(StartInfo);
    StartInfo.lpDesktop := PChar('');
    StartInfo.dwX := 10000;
    StartInfo.dwY := 10000;
	StartInfo.dwFlags := STARTF_USEPOSITION;
    SetErrorMode(SEM_FAILCRITICALERRORS);
    if not CreateProcess(NIL, PChar(Executable), NIL, NIL, FALSE,
                         CREATE_NO_WINDOW, NIL, PChar(ExecFolder), StartInfo,
                         ProcessInfo) then
    begin
    	WriteLn(f, TestNum, ')  0 -    ');
        CloseFile(f);
        Exit;
    end;

    R := TestProcess(ProcessInfo.hProcess, TimeLimit, MemoryLimit);

    CloseHandle(ProcessInfo.hProcess);
    CloseHandle(ProcessInfo.hThread);

    if R <> '' then
    begin
    	WriteLn(f, TestNum, ')  0 - ', R);
        CloseFile(f);
        Exit;
    end;

    { Launch checker }
    {
    CallLine := TaskFolder + 'check.exe ' + IntToStr(maxScore) + ' ' +
                TaskFolder + TestNum + '.in ' +
                TaskName + '.out ' +
                TaskFolder + TestNum + '.ans';
     }
     CallLine := TaskFolder + 'check.bat ' + IntToStr(maxScore) + ' ' +
                TaskFolder + TestNum + '.in ' +
                TaskName + '.out ' +
                TaskFolder + TestNum + '.ans';

    if not CreateProcess(NIL, PChar(CallLine), NIL, NIL, FALSE,
                         CREATE_NO_WINDOW, NIL, PChar(ExecFolder), StartInfo,
                         ProcessInfo) then
    begin
    	WriteLn(f, TestNum, ')  ? -     ');
        CloseFile(f);
        Exit;
    end;

    WaitForSingleObject(ProcessInfo.hProcess, INFINITE);

    CloseHandle(ProcessInfo.hProcess);
    CloseHandle(ProcessInfo.hThread);

    AssignFile(fRes, ExecFolder + 'report');
{$I-}
    Reset(fRes);
    if IOResult = 0 then
    begin
    	ReadLn(fRes, R);
      ttt := Pos('  ',R);
      if Pos('  ',R) <> 0 then

        R := Copy(R,1,Pos('  ',R)-1) +
                            '  ' +
             Copy(R,Pos('  ',R)
             +Length('  '),
             Length(R));
	    CloseFile(fRes);
      while (R<>'') and (R[1]=' ') do delete(R,1,1);
        if R[1] = ' ' then
        	R[1] := '0';
        for i := 1 to Length(R) do
        	if (R[i] <= '9') and (R[i] >= '0') then
            	Result := Result * 10 + ord(R[i]) - ord('0') else
                break;

        WriteLn(f, TestNum, ') ', R);
    end else
    begin
    	WriteLn(f, TestNum, ')  ? -     ');
        CloseFile(f);
        Exit;
    end;
{$I+}

    CloseFile(f);
end;

{ DoGrading()
  Compile, read test description, etc }
procedure DoGrading(const Job: TJob);
var
    FullLog, ShortLog, FileName, NewName, Executable: String;

    i, ind, total: Integer;
    TaskDescrFile: String;
    f: TextFile;
    nTest, TimeLimit, MemoryLimit: Integer;
    Score: array of Integer;
begin
	indGrade := Job.Index;
    
	{ Find correct file for this task }
  repeat
	FileName := FindTask(Job.Folder, Job.Task);
    if FileName = '' then
    begin
    	MakeReport(Job.Folder, Job.Task, '³ ''',
        	'   ''    ' + Job.Task);
        break;
    end;

    { Copy solution to temporary directory }
    NewName := WorkDir + '\' + TEST_DIR + '\a.' + GetExt(FileName);
    if not CopyFile(PChar(FileName), PChar(NewName), False) then
    begin
    	MakeReport(Job.Folder, Job.Task, ' !',
        	'   ''   !');
        Break;
    end else
    	FileName := NewName;

    { Compile solution }
    FullLog := Job.Folder + Job.Task + FULL_LOG_SUFFIX;
    ShortLog := Job.Folder + Job.Task + SHORT_LOG_SUFFIX;
    Executable := Compile(FileName, FullLog);
    if Executable = '' then
    begin
    	MakeReport(Job.Folder, Job.Task, ' ', '');
        Break;
    end;

    { Read task description }
    TaskDescrFile := '';
    ind := 0;
    for i := 0 to Length( TaskList )-1 do
        if TaskList[i].Name = Job.Task then
        begin
        	ind := i;
        	TaskDescrFile := TaskList[i].Folder + 'task.cfg';
            Break;
        end;

    AssignFile(f, TaskDescrFile);
    Reset(f);
    	ReadLn(f, nTest, TimeLimit, MemoryLimit);
        SetLength(Score, nTest);
        for i := 0 to nTest-1 do
        	Read(f, Score[i]);
    CloseFile(f);

    { Execute program itself }
    AssignFile(f, FullLog);
    Rewrite(f);
    Close(f);

    total := 0;
    for i := 1 to nTest do
    	Inc(total, DoTest(Executable, FullLog, Job.Task, TaskList[ind].Folder,
		                  i, Score[i-1], TimeLimit, MemoryLimit*1024));

    AssignFile(f, ShortLog);
    Rewrite(f);
    	WriteLn(f, total);
    Close(f);
	SetLength(UpdateList, Length(UpdateList)+1);
    UpdateList[ Length(UpdateList)-1 ] := Job.Index;
  until true;
end;

{ GradeThreadProc()
  Grading thread procedure }
procedure GradeThreadProc(lpData: Pointer);
var
	Job: TJob;
begin
	while TRUE do
    begin
    	if (StopGrading) OR (NOT PopJob(Job)) then
        	Break else
	        DoGrading(Job);
    end;
	GradeThreadExec := False;
    ExitThread(0);
end;

end.

